use std::collections::HashMap;

use futures::prelude::*;
use htmlescape::decode_html;
use irc::client::prelude::*;
use onig::*;
use rspotify::{ClientCredsSpotify, Credentials};
use rspotify::model::{Id, PlayableItem};
use rspotify::prelude::BaseClient;

fn calculate_playtime(secs: u64) -> (u64, u64) {
    let mut dur_sec = secs;
    let dur_min = dur_sec / 60;
    dur_sec -= dur_min * 60;
    (dur_min, dur_sec)
}

async fn resolve_spotify(spotify: &ClientCredsSpotify, resource_type: &str, resource_id: &str) -> anyhow::Result<String> {
    match resource_type {
        "track" => {
            let track = spotify.track(Id::from_id(resource_id)?).await?;
            let playtime = calculate_playtime(track.duration.as_secs());
            let artists: Vec<String> = track.artists.into_iter().map(|x| x.name).collect();
            Ok(format!("\x037[Spotify]\x03 Track: \x039\"{}\"\x03 - \x039\"{}\" \x0311|\x03 Album: \x039\"{}\" \x0311|\x03 Length:\x0315 {}:{:02} \x0311|", artists.join(", "), track.name, track.album.name, playtime.0, playtime.1))
        }
        "artist" => {
            let artist = spotify.artist(Id::from_id(resource_id)?).await?;
            Ok(format!("\x037[Spotify]\x03 Artist: \x039\"{}\" \x0311|\x03 Genres:\x039 {} \x0311|", artist.name, artist.genres.join(", ")))
        }
        "album" => {
            let album = spotify.album(Id::from_id(resource_id)?).await?;
            let playtime = calculate_playtime(album.tracks.items.iter().fold(0, |acc, x| acc+x.duration.as_secs()));
            Ok(format!("\x037[Spotify]\x03 Album: \x039\"{}\" \x0311|\x03 Tracks:\x0315 {} \x0311|\x03 Release date:\x039 {} \x0311|\x03 Length:\x0315 {}:{:02} \x0311|", album.name, album.tracks.total, album.release_date, playtime.0, playtime.1))
        }
        "playlist" => {
            let playlist = spotify.playlist(Id::from_id(resource_id)?, None, None).await?;
            let mut tracks = 0;
            let playtime = calculate_playtime(playlist.tracks.items.iter().fold(0, |acc, x| {
                if let Some(item) = &x.track {
                    match item {
                        PlayableItem::Track(t) => {
                            tracks += 1;
                            acc+t.duration.as_secs()
                        }
                        PlayableItem::Episode(e) => {
                            tracks += 1;
                            acc+e.duration.as_secs()
                        }
                    }
                } else {
                    acc
                }
            }));
            Ok(format!("\x037[Spotify]\x03 Playlist: \x039\"{}\" \x0311|\x03 Tracks/Episodes:\x0315 {} \x0311|\x03 Length:\x0315 {}:{:02} \x0311|\x03 Description: \x039\"{}\" \x0311|", playlist.name, tracks, playtime.0, playtime.1, playlist.description.unwrap_or("<empty>".into())))
        }
        _ => Ok("\x037[Spotify]\x03 Error: Invalid resource type".into())
    }
}

#[tokio::main]
async fn main() -> anyhow::Result<()> {
    if let Some(filename) = std::env::args().nth(1) {
        let config = Config::load(filename)?;
        let mut spotify = ClientCredsSpotify::new(Credentials::from_env().unwrap());
        spotify.request_token().await.unwrap();
        let mut client = Client::from_config(config).await?;
        client.identify()?;

        println!("Connected to IRC!");

        let mut stream = client.stream()?;

        let url_regex = Regex::new(r"https?://\w+\.\w+[/\S+]*").unwrap();
        let title_regex = Regex::new(r"(?<=<title>)(.*)(?=</title>)").unwrap();
        let spotify_regex = Regex::new(r"(?:https?|spotify):(?://open\.spotify\.com/)?(track|artist|album|playlist)[/:]([a-zA-Z0-9]*)").unwrap();

        let mut cache: HashMap<String, String> = HashMap::new();

        while let Some(message) = stream.next().await.transpose()? {
            print!("{}", message);

            if let Command::PRIVMSG(channel, message) = message.command {
                if let Some(m) = spotify_regex.captures(&message) {
                    let tp_group = m.pos(1).unwrap();
                    let id_group = m.pos(2).unwrap();
                    match resolve_spotify(&spotify, &message[tp_group.0..tp_group.1], &message[id_group.0..id_group.1]).await {
                        Ok(o) => client.send_privmsg(&channel, o)?,
                        Err(_) => client.send_privmsg(&channel, "\x037[Spotify]\x03 An error has occurred.")?
                    }
                } else if let Some(m) = url_regex.find(&message) {
                    let url = &message[m.0..m.1];
                    if let Some(entry) = cache.get(&url.to_string()) {
                        client.send_privmsg(&channel, format!("\x039[Title]\x03 {}", entry))?;
                        continue;
                    }
                    if let Ok(o) = reqwest::get(url).await {
                        let body = o.text().await?;
                        if let Some(tm) = title_regex.find(&body) {
                            let title_match = &body[tm.0..tm.1];
                            let result = decode_html(&title_match).unwrap_or(title_match.to_string());
                            client.send_privmsg(&channel, format!("\x039[Title]\x03 {}", result))?;
                            cache.insert(url.to_string(), result.to_string());
                            continue;
                        }
                    }
                    client.send_privmsg(&channel, "\x039[Title]\x03 Could not get the title.")?;
                }
            }
        }
    } else {
        println!("Usage: titlebot <path to config.toml>")
    }
    Ok(())
}
